-- Talky NPC, which speaks without interactions (WIP)

require "data_talks.npc_talks"

Modules.Talky = true

canTalk = true -- Can this NPC start a talk right now
talkName = ""
talkMembers = {} -- Current talk members
acceptedInvites = 0 -- Accepted talk invites
reqInvites = 0

Talk = {}

-- Quotes from NPC
--[[Quick example
Quotes = {
	Situation1 = { -- See ai_modular.lua for situation - basically it is just a string set by scripting
		"Talk1",
		"Talk2",
		...
	},

	Default = {
		"Default talk here"
	}
}
]]
Talk.Quotes = {}

-- Random speech, can include paths or quotes;
--[[Example
RandomSpeech = {
	Situation = {
		QuoteMin = 1, -- Minimal quote delay, in seconds
		QuoteMax = 10, -- Maximal quote delay, in seconds
	}
}
]]
Talk.RandomSpeech = {}

-- Greeting messages
--[[Example
GreetMessages = {
	{
		Situation = "",
		Conditions = "",
		Greeting = "Hello, {player}!",
		Farewell = "See you!"
	}
}
]]
Talk.GreetMessages = {}
Talk.GreetPlayers = false
Talk.GreetDistance = 8

-- Special functions - can be used as talk paths (current talk name argument)
TalkHandles = {}

-- Takes and displays random quote from quote table
function RandomQuote()
	local relevant = Talk.Quotes[Situation] -- Relevant quotes to this situation
	if relevant == nil or #relevant == 0 then return end -- No quotes?

	local index = math.random(1, #relevant)
	this:NpcSpeech(relevant[index]) -- Speak the line!
end

-- Initializes a talk with given name using given NPCs
function StartTalk(name, members, force)
	--DebugMessage("Starting talk!")
	if not canTalk and not force then return end
	--DebugMessage("Can talk right now")
	talkMembers = members

	local membersCount = 0
	for k,v in pairs(members) do
		membersCount = membersCount + 1
	end
	reqInvites = membersCount

	DebugTable(talkMembers)

	for role,obj in pairs(members) do
		obj:SendMessage("InviteTalk", this, name, members)
	end
end

-- Continue current talk from given position
function ContinueTalk(name, pos)
	if name ~= talkName then return end
	--DebugMessage("Continue talk now!")

	local data = Discussions[talkName]
	local node = data[pos]
	local speaker = node[1] -- Correct speaker

	if talkMembers[speaker] == this then
		--DebugMessage("Corrent talker found, begin initializing talk!")
		local text = node[2]
		local handle = TalkHandles[text]
		if handle ~= nil then
			--DebugMessage("Execute talk handle")
			handle(talkName)
		else
			DebugMessage("Execute NPC speech")
			this:NpcSpeech(text)
		end

		if pos == #data then
			--DebugMessage("Talk ended naturally after last line")
			EndTalk()
			return
		end
		RegisterSingleEventHandler(EventType.Timer, "WaitTalkDelay", function ()
			--DebugMessage("After waiting talk delay, next quote is said")
			local nextNode = data[pos + 1]
			local nextSpeaker = nextNode[1]
			local obj = talkMembers[nextSpeaker]

			obj:SendMessage("ContinueTalk", talkName, pos + 1)
		end)
		this:ScheduleTimerDelay(TimeSpan.FromSeconds(node[3] or 4), "WaitTalkDelay")
	else -- Redirect to correct speaker
		--DebugMessage("Redirecting talk to another NPC")
		talkMembers[speaker]:SendMessage("ContinueTalk", talkName, pos)
	end
end

-- End talk by sending clear order to all members (make sure this is current speaker when calling)
function EndTalk()
	for role,obj in pairs(talkMembers) do
		obj:SendMessage("ClearTalk", talkName)
	end
end

-- Clear talk data of this NPC
function ClearTalk(name)
	if name ~= talkName then return end -- Not relevant order

	canTalk = true
	talkName = ""
	talkMembers = {}
	acceptedInvites = 0
	reqInvites = 0

	if TalkHandles.OnFinish ~= nil then
		TalkHandles.OnFinish(name)
	end
end

-- Handle other object's response of invite
function HandleInviteResponse(result, name)
	--DebugMessage("Got talk invite response!")
	if not result and talkName == name then
		--DebugMessage("Talk ended, could not get all members")
		EndTalk()
	elseif name == talkName then
		acceptedInvites = acceptedInvites + 1
		--DebugMessage("Accepted invite number "..acceptedInvites.." of "..reqInvites)
		--DebugTable(talkMembers)

		if acceptedInvites >= membersCount then
			--DebugMessage("Got all members, starting talk!")
			ContinueTalk(name, 1) -- At beginning
		end
	end
end

-- Handles and responds to talk invite
function HandleTalkInvite(source, name, members)
	--DebugMessage("Got talk invite!")
	local response
	if Talk.CustomInviteHandle ~= nil then
		response = Talk.CustomInviteHandle(source, name, members)
	else
		response = canTalk
	end

	if response then
		canTalk = false
		talkName = name
		talkMembers = members
	end
	--DebugMessage("Response is "..tostring(response))

	source:SendMessage("InvTalkResponse", response, name)
end

function GreetPlayer(player)
	local displayName = StripColorFromString(player:GetName())
	for i,msg in ipairs(Talk.GreetMessages) do
		if CheckConditions(msg.Conditions, player) and msg.Situation == Situation then
			local text = msg.Greeting
			if not text then return end
			text = text:gsub("{player}", displayName)
			this:NpcSpeech(text)
			return
		end
	end
end

function FarewellPlayer(player)
	local displayName = StripColorFromString(player:GetName())
	for i,msg in ipairs(Talk.GreetMessages) do
		if CheckConditions(msg.Conditions, player) and msg.Situation == Situation then
			text = msg.Farewell
			if not text then return end
			text = text:gsub("{player}", displayName)
			this:NpcSpeech(text)
			return
		end
	end
end

function UpdateTalkFlags()
	local hasView = HasView("GreetPlayers")

	if Talk.GreetPlayers and not hasView then
		AddView("GreetPlayers", SearchPlayerInRange(Talk.GreetDistance, false))
	elseif not Talk.GreetPlayers and hasView then
		DelView("GreetPlayers")
	end
end

if not Modules.Dialogue then -- If yes, use dialogue data handling
	if initializer ~= nil and initializer.DialogueData ~= nil then
		this:SetObjVar("DialogueData", initializer.DialogueData)
	end

	if this:HasObjVar("DialogueData") then
		AddPostRequire("data_dialogue."..this:GetObjVar("DialogueData"))
	end
end

RegisterEventHandler(EventType.EnterView, "GreetPlayers", GreetPlayer)
RegisterEventHandler(EventType.LeaveView, "GreetPlayers", FarewellPlayer)

RegisterEventHandler(EventType.Message, "InviteTalk", HandleTalkInvite)
RegisterEventHandler(EventType.Message, "InvTalkResponse", HandleInviteResponse)
RegisterEventHandler(EventType.Message, "ContinueTalk", ContinueTalk)
RegisterEventHandler(EventType.Message, "StartTalk", StartTalk)

AddPostTask(UpdateTalkFlags)