require 'incl_container'
require 'incl_equipmentstats'
require 'incl_dialogwindow'
require 'base_mobile'

nextTransactionIndex = 0
activeTransactions = {}

Merchant = {
	CurrencyInfo =
	{
		CurrencyType = "coins",
		ObjVarName = nil,
		CurrencyDisplayStr = "coins"
	}
}

function ValidateTransaction(curState, transactionId)
	local transactionData = activeTransactions[transactionId]

	-- transaction in wrong state
	if( transactionData == nil or transactionData.State ~= curState ) then		
		return false
	end	

	-- buyer no longer valid
	local buyer = transactionData.Buyer
	if( buyer == nil or not(buyer:IsValid()) ) then		
		return false
	end
	
	local item = transactionData.Item
	-- something happened to the display item
	if( item == nil or not(item:IsValid()) ) then	
		return false
	end

	return true
end

function GetModifiedBuyerPrice(itemPrice, buyer)
	--D*ebugMessage("b" .. tostring(buyer))
	local buyerPriceMod = buyer:GetObjVar("merchantReputation") or 1
	if(buyerPriceMod > 0) then
		buyerPriceMod =100 - (buyerPriceMod / 4)
	else
		buyerPriceMod = 100 - (buyerPriceMod * 2.5)
	end
	
	itemPrice = (itemPrice * buyerPriceMod) / 100
	return math.ceil(itemPrice)

end

function GetModifiedSellValue(itemValue,seller)
	local sellerPriceMod = seller:GetObjVar("merchantReputation") or 1
	if(sellerPriceMod > 0) then
		sellerPriceMod =100 - (sellerPriceMod / 4)
	else
		sellerPriceMod = 100 + (sellerPriceMod * .8)
	end
	
	itemValue = (itemValue * sellerPriceMod) / 100
	return math.ceil(itemValue)
end

function CanUserSell(item,user)
	if( item:TopmostContainer() ~= user ) then		
		return false
	end	

	return true
end

function GetNumBonuses(item)
	local objVarDict = item:GetAllObjVars()

	local bonusCount = 0
	for name,varValue in pairs(objVarDict) do
		if( name:match("Bonus") ) then
			bonusCount = bonusCount + 1
		end
	end

	return bonusCount
end

customSellValues = 
{
	-- no buying back
	dye = 0,
	dye_leather = 0,
	dye_cloth = 0,
	dye_metal = 0,
	dye_weapon = 0,
	dye_skin = 0,
}


function GetItemSellValue(item)	
	-- first check for custom value by template id
	local itemTemplateId = item:GetCreationTemplateId()
	if( customSellValues[itemTemplateId] ~= nil ) then
		return customSellValues[itemTemplateId]
	end

	-- most animal parts are worth 5
	if( itemTemplateId:match("animalparts") ) then
		return 3
	end

	if (item:HasObjVar("Worthless")) then
		return 0 
	end
	if (item:HasObjVar("Valuable")) then
		return item:GetObjVar("Valuable")
	end

	-- some resource types have value
	local resourceType = item:GetObjVar("ResourceType")
	if( resourceType ~= nil ) then
		if( resourceType == "coins") then
			return 0
		-- refresh potions are worth more
		elseif( resourceType == "potion_refresh_regen") then
			return 5		
		elseif( resourceType:match("potion") ) then
			return 3
		elseif( resourceType:match("bandages") ) then
			return 3
		elseif( resourceType:match("Wood") ) then
			return 1
		elseif( resourceType:match("Stone") ) then
			return 1
		elseif( resourceType:match("Iron") ) then
			return 3
		end
	end

	local foodValue = item:GetObjVar("FoodValue")
	if( foodValue ~= nil ) then
		return math.max(1,math.floor(foodValue/4))
	end

	-- scrolls and skill books
	if( item:HasModule("spell_scroll") or item:HasModule("skill_book") ) then
		return 5
	end

	local equipClass = GetEquipmentClass(item)
	if( equipClass == "WeaponClass" ) then
		local baseValue = 0
		local weaponClass = GetWeaponClass(item)
		if( weaponClass == "Vshort" ) then
			baseValue = 2
		elseif( weaponClass == "Short" or weaponClass == "ShortMid" ) then
			baseValue = 3
		elseif( weaponClass == "Mid" or weaponClass == "MidLong" ) then
			baseValue = 4
		elseif( weaponClass == "Long" ) then
			baseValue = 5
		end

		return baseValue + (GetNumBonuses(item) * 2)
	elseif( equipClass == "ArmorClass" ) then
		local baseValue = 0
		local baseMult = 1
		local armorClass = GetArmorClass(item)
		-- padded
		if( armorClass == "VeryLight") then
			baseMult = 1
		-- leather
		elseif( armorClass == "Light") then
			baseMult = 1.5
		-- chain and scale
		elseif( armorClass == "Medium" or armorClass == "Heavy") then
			baseMult = 2
		-- plate
		elseif( armorClass == "VeryHeavy") then
			baseMult = 2.5
		end

		local equipSlot = GetEquipSlot(item)
		if( equipSlot == "Head" or equipSlot == "Legs") then
			baseValue = 2 * baseMult
		elseif( equipSlot == "Chest" ) then
			baseValue = 4 * baseMult		
		end

		local combinedValue = baseValue + (GetNumBonuses(item) * 2)

		-- cultist armor is worth half
		if(itemTemplateId:match("cultist")) then
			combinedValue = combinedValue / 2
		end

		return math.floor(combinedValue)

	elseif( equipClass == "ShieldClass" ) then
		local baseValue = 0
		local shieldClass = GetShieldClass(item)
		if( shieldClass == "Buckler") then
			baseValue = 2
		elseif( shieldClass == "Small" or shieldClass == "Medium") then
			baseValue = 3
		elseif( shieldClass == "Large" or armorClass == "ExtraLarge") then
			baseValue = 5		
		end

		return baseValue + (GetNumBonuses(item) * 2)
	end

	return 0
end

function ValidateSellItem(item,user)
	if( not(CanUserSell(item,user)) ) then
    	this:NpcSpeech("That is not yours to sell!")
    	return 0
    end
    local sellValue = GetItemSellValue(item)
    if( sellValue == 0 ) then
    	this:NpcSpeech("I'm not interested in that.")
    	return 0
    end

    local stackCount = item:GetStackCount()

    return sellValue * stackCount
end

function GiveCoins(user,amount)
	local backpackObj = user:GetEquippedObject("Backpack")
	if( backpackObj ~= nil ) then
		-- todo: make an easier way to create stacks
    	local packObjects = backpackObj:GetContainedObjects()
     	for index, packObj in pairs(packObjects) do	
     		if( packObj:GetObjVar("ResourceType") == "coins" ) then
     			packObj:RequestAddToStack(amount)
     			return
     		end
     	end

    	CreateObjInBackpack(user,"coin_purse","coins",amount)
    end
end

function DoPurchase(transactionId)
	local transactionData = activeTransactions[transactionId]
	local buyer = transactionData.Buyer
	local item = transactionData.Item
	local itemTemplate = item:GetCreationTemplateId()
	local createSuccess = false

	local backpackObj = buyer:GetEquippedObject("Backpack")
	-- TODO: Verify creation success and refund money on failure
	if( backpackObj ~= nil ) then			
		-- first try to stack
		local resourceType = item:GetObjVar("ResourceType")
		if( TryAddToStack(resourceType, backpackObj, 1) ) then			
			createSuccess = true
		else
			-- if it didnt stack, create in container
			local randomLoc = GetRandomDropPosition(this)
			-- create the dye
			
			CreationData={
				DyeColor	=transactionData.Item:GetObjVar("DyeColor"),
				Uses 		=transactionData.Item:GetObjVar("Uses"),
				SatCost 	=transactionData.Item:GetObjVar("SatCost"),
				DyeType 	=transactionData.Item:GetObjVar("DyeType"),
				NamePreFix 	=transactionData.Item:GetObjVar("NamePreFix"),
				NamePostFix =transactionData.Item:GetObjVar("NamePostFix"),
				}
		
			CreateObjInContainer(itemTemplate, backpackObj, randomLoc, "DyeCreated", CreationData)

			createSuccess = true;
		end		
	else
		-- should we really put this thing on the ground?
		createSuccess = CreateObj(itemTemplate, buyer:GetLoc())
	end

	if( createSuccess ) then
		Merchant.DoPurchaseComplete(buyer)
	end	
end

function Merchant.DoAppraise(user)
	-- start the appraisal process
	user:RequestClientTargetGameObj(this, "merchant_appraise")
end

function Merchant.DoSell(user)
	-- start the sale process
	user:SystemMessage("Click on the item you wish to sell. (ESC to cancel)")
	user:RequestClientTargetGameObj(this, "merchant_sell")
end

function Merchant.DoCantAfford(buyer)
	Speak{Text="I beg your pardon "..buyer:GetName().." but I don't believe you can afford that."}
end

function Merchant.DoPurchaseComplete(buyer)
	local myRep = buyer:GetObjVar("merchantReputation") or 0
    myRep = myRep + .1
    buyer:SetObjVar("merchantReputation", myRep)
	Speak{Text="Thank you for your patronage "..buyer:GetName()..".", Audio="MerchantPlayerPurchase"}
end

function HandleSaleItemCreated(success,objRef,itemEntry)	
	if(success) then
		-- Dye initialization
			objRef:SetObjVar("DyeColor"		,itemEntry.DyeColor   )
			objRef:SetObjVar("Uses"			,itemEntry.Uses 	  )
			objRef:SetObjVar("SatCost"		,itemEntry.SatCost	  )
			objRef:SetObjVar("DyeType"		,itemEntry.DyeType 	  )
			objRef:SetObjVar("NamePreFix"	,itemEntry.NamePreFix )
			objRef:SetObjVar("NamePostFix"	,itemEntry.NamePostFix)

			objRef:SendMessage("Adjust",itemEntry.DyeColor)

			DebugMessage(
				    "itemEntry.Price: "	.. tostring(itemEntry.Price)
				..", itemEntry.DyeColor: "		.. tostring(itemEntry.DyeColor )
				..", itemEntry.Uses: "			.. tostring(itemEntry.Uses)
				..", itemEntry.SatCost: "		.. tostring(itemEntry.SatCost)
				..", itemEntry.DyeType: "		.. tostring(itemEntry.DyeType)
				..", itemEntry.NamePreFix: "	.. tostring(itemEntry.NamePreFix )
				..", itemEntry.NamePostFix: "	.. tostring(itemEntry.NamePostFix)
				)

		-- Now the merchant stuff
		objRef:AddModule("merchant_sale_item")
		objRef:SendMessage("InitSaleItem",{ tonumber(itemEntry.Price),this})
		

		
	end
end

function CanBuyItem(buyer,item)
	return true
end

function HandleSellItem(args)
	local buyer,item = table.unpack(args)

	if( buyer == nil or not(buyer:IsValid()) ) then return end
	if( item == nil or not(item:IsValid()) ) then return end
	if (not CanBuyItem(buyer,item)) then return end
	
	local itemTemplate = item:GetCreationTemplateId()
	local itemPrice = item:GetObjVar("itemPrice") or 0	
	
	itemPrice = GetModifiedBuyerPrice(itemPrice, buyer)

	--D*ebugMessage("ModPrice:" .. tostring(itemPrice))	
	local body = "Do you wish to purchase the "..item:GetName().." for "..itemPrice.."?"
	local transactionId = "Transaction"..nextTransactionIndex

	ClientDialog.Show{
	    TargetUser = buyer,
	    DialogId = transactionId,
	    TitleStr = "Confirm Purchase",
	    DescStr = body,
	    Button1Str = "Confirm",
	    Button2Str = "Cancel",
	    ResponseObj = this,
	}

	activeTransactions[transactionId] = { State="Confirm", Item=item, Buyer = buyer }

	nextTransactionIndex = nextTransactionIndex + 1
end

function HandleConsumeResourceResponse(args)
	local success,transactionId,buyer = table.unpack(args)

	if( not(ValidateTransaction("TakeCoins",transactionId))) then
		activeTransactions[transactionId] = nil
		return
	end

	-- something went wrong taking the coins
	if( not(success) ) then
		Merchant.DoCantAfford(buyer)		
	-- we have taken the coins complete the transaction
	else
		DoPurchase(transactionId)		
	end

	activeTransactions[transactionId] = nil
end

function HandleDialogResult(user, transactionId, buttonId)	
	-- id 1 is cancel
	if( buttonId == 1 ) then
		activeTransactions[transactionId] = nil
		return
	end

	if( not(ValidateTransaction("Confirm",transactionId))) then
		activeTransactions[transactionId] = nil
		return
	end

	local transactionData = activeTransactions[transactionId]

	-- make sure they can afford it
	local item = transactionData.Item
	local buyer = transactionData.Buyer
	local amount = item:GetObjVar("itemPrice")
	amount = GetModifiedBuyerPrice(amount, buyer)
	if( Merchant.CurrencyInfo.CurrencyType == "coins" ) then
		if( buyer:CountCoins() < amount ) then
			Merchant.DoCantAfford(buyer)
			activeTransactions[transactionId] = nil
			return
		else
			-- everything checks out lets take the coins
			transactionData.State = "TakeCoins"
			buyer:RequestConsumeResource("coins",amount,transactionId,this)
		end
	elseif( Merchant.CurrencyInfo.CurrencyType == "objvar" ) then
		curValue = buyer:GetObjVar(Merchant.CurrencyInfo.ObjVarName) or 0
		if( curValue < amount ) then
			Merchant.DoCantAfford(buyer)
			activeTransactions[transactionId] = nil
			return
		else
			-- everything checks out 
			buyer:SetObjVar(Merchant.CurrencyInfo.ObjVarName, curValue - amount)
			DoPurchase(transactionId)
		end
	end
end

function HandleAppraiseTargeted(target,user)
	if( target == nil ) then return end
    if( user == nil or not(user:IsValid())) then return end

    local sellValue = ValidateSellItem(target,user)
    sellValue = GetModifiedSellValue(sellValue, user)
    if( sellValue > 0 ) then
    	local currency = "coins"
    	this:NpcSpeech("I'll give you "..sellValue.." "..Merchant.CurrencyInfo.CurrencyDisplayStr.." for the "..target:GetName())
    end
end

function HandleSellTargeted(target,user)
	if( target == nil ) then return end
    if( user == nil or not(user:IsValid())) then return end

    local sellValue = GetItemSellValue(target)
    --D*ebugMessage("SellVal:  ".. tostring(sellValue))
    sellValue = GetModifiedSellValue(sellValue,user)
    --D*ebugMessage("ModSellVal: " .. tostring(sellValue))
    local sellValue = ValidateSellItem(target,user)
    if( sellValue > 0 ) then
    	target:Destroy()
    	
    	-- TODO: Handle other currencies
    	GiveCoins(user,sellValue)
    	local myRep = user:GetObjVar("merchantReputation") or 0
    	myRep = myRep + .1
    	user:SetObjVar("merchantReputation", myRep)
    	Speak{Text="Here is your "..sellValue.." "..Merchant.CurrencyInfo.CurrencyDisplayStr..". Thank you for your patronage "..user:GetName()..".", Audio="MerchantPlayerPurchase"}    	
    end

    Merchant.DoSell(user)
end

RegisterEventHandler(EventType.CreatedObject, "coins", 
	function(success,objRef,amount)
		if(success and amount > 1) then
			objRef:RequestAddToStack(amount-1)
		end
	end
)

function HandleDyeCreated( success, objRef, creationData )
	if success then 
		DebugMessage("Dye Created - success ! Now adjusting Item")
		DebugMessage(IdStr(objRef)
		.."creationData: "
		.."DyeColor: ".. 		tostring(creationData.DyeColor)
		..", Uses: ".. 			tostring(creationData.Uses)
		..", SatCost: ".. 		tostring(creationData.SatCost)
		..", DyeType: ".. 		tostring(creationData.DyeType)
		..", NamePreFix: "..	tostring(creationData.NamePreFix)
		..", NamePostFix: ".. 	tostring(creationData.NamePostFix)
			)


		objRef:SetObjVar("DyeColor", creationData.DyeColor )
		objRef:SetObjVar("Uses", creationData.Uses)
		objRef:SetObjVar("SatCost", creationData.SatCost)
		objRef:SetObjVar("DyeType", creationData.DyeType)
		objRef:SetObjVar("NamePreFix", creationData.NamePreFix)
		objRef:SetObjVar("NamePostFix", creationData.NamePostFix)
		
		objRef:SendMessage("Adjust",creationData.DyeColor)
	end

end
--Debug Helper
	function IdStr(target)
		return tostring(target:GetName())..":"..tostring(target.Id)
	end


RegisterEventHandler(EventType.CreatedObject, "sale_item_created", HandleSaleItemCreated)
RegisterEventHandler(EventType.Message,"SellItem", HandleSellItem)
RegisterEventHandler(EventType.Message, "ConsumeResourceResponse", HandleConsumeResourceResponse)
RegisterEventHandler(EventType.ClientObjectCommand,"dialogResult", HandleDialogResult)
RegisterEventHandler(EventType.ClientTargetGameObjResponse, "merchant_appraise", HandleAppraiseTargeted)
RegisterEventHandler(EventType.ClientTargetGameObjResponse, "merchant_sell", HandleSellTargeted)

RegisterEventHandler(EventType.CreatedObject,"DyeCreated",HandleDyeCreated)

-- first time setup
if(initializer ~= nil) then
	local merchantContainers = {}
	if initializer.MerchantContainers ~= nil then
		for index, itemEntry in pairs(initializer.MerchantContainers) do
			local findResult = FindObjects(SearchMulti(
	      		{
	      			SearchObjVar("merchantContainer",itemEntry.Name),
					SearchObjectInRange(10),
	  			}))

			if( #findResult > 0 ) then				
				local contObj = findResult[1]

				merchantContainers[itemEntry.Name] = contObj

				if(itemEntry.DisplayName ~= nil) then
					contObj:SetName(itemEntry.DisplayName)
				end

				if(itemEntry.TooltipStr ~= nil) then
					contObj:SetSharedObjectProperty("TooltipString",itemEntry.TooltipStr)
				end
			end
		end
	end

	for index, itemEntry in pairs(initializer.ItemInventory) do
		local itemLoc = ParseLoc(itemEntry.Loc)
		local itemRot = Loc(0,0,0)
		local itemScale = Loc(1,1,1)
		if( itemEntry.Rotation ~= nil ) then
			itemRot = ParseLoc(itemEntry.Rotation)
		end

		if( itemLoc ~= nil) then
			local contObj = nil
			if( itemEntry.Container ~= nil and merchantContainers[itemEntry.Container] ~= nil) then
				contObj = merchantContainers[itemEntry.Container]				
			end

			if( itemEntry.Container ~= nil and contObj == nil ) then
				DebugMessage("[Merchant:Init] Merchant container not found for item "..itemEntry.Template)
			else
				--DebugMessage("itemEntry.Template: "..tostring(itemEntry.Template).." Location:"..tostring(itemLoc))
				CreateObjExtended(itemEntry.Template,contObj,itemLoc,itemRot,itemScale,"sale_item_created", itemEntry)
			end
		end
	end
end
